module Calendar.Gregorian where

import Calendar.Global

gregorianEpoch = 1

data Gregorian = GregorianData {gregorian'year::Fixed, gregorian'month::Fixed,
                                             gregorian'day::Fixed} deriving (Eq)

toGregorian :: Fixed -> Fixed -> Fixed -> Gregorian
toGregorian year month day = GregorianData year month day

zeropad thestring thelength =
    if (length thestring) >= thelength then
        thestring -- Which means we may return a string longer than thelength
    else -- Pad with leading zeroes
        (replicate (thelength - length thestring) '0') ++ thestring

-- We do not derive Show because ours is better (ISO 8601 look)
instance Show Gregorian where
    show = \g -> zeropad (show (gregorian'year g)) 4 ++ "-" ++
           zeropad (show (gregorian'month g)) 2 ++ "-" ++
           zeropad (show (gregorian'day g)) 2 

-- Default Eq is fine, we do not need to define it. If we had to, it would be:
--instance Eq Gregorian where
--    (==) = \g1 -> \g2 -> gregorian'year g1 == gregorian'year g2 ...

prettyShow g =
    show (gregorian'day g) ++ " " ++
    case(gregorian'month g) of
               1 -> "January"
               2 -> "February"      
               3 -> "March"
               4 -> "April"
               5 -> "May"
               6 -> "June"
               7 -> "July"
               8 -> "August"
               9 -> "September"
               10 -> "October"
               11 -> "November"
               12 -> "December"
               otherwise -> error "Month out of range"
          ++ " " ++
     show (gregorian'year g) 

gregorianLeapYear y =
   (y `mod` 4) == 0 &&
               ((y `mod` 100) /= 0 || y `mod` 400 == 0)

gregorianOffset year month =
    if month <= 2 then 0
    else if gregorianLeapYear year then -1
         else -2

fixedFromGregorian :: Gregorian -> Fixed
fixedFromGregorian gd = 
    let year = gregorian'year gd in
    let month = gregorian'month gd in
    let day = gregorian'day gd in
       gregorianEpoch - 1 + 365 * (year - 1) +
                   floor (fromIntegral (year - 1) / 4) -
                   floor (fromIntegral (year - 1) / 100) +
                   floor (fromIntegral (year - 1) / 400) +
                   floor (fromIntegral (367 * month - 362) / 12) +
                   gregorianOffset year month +
                   day

gregorianYearFromFixed fd =
    let d0 = fd - gregorianEpoch in
    let n400 = floor (fromIntegral d0 / 146097) in
    let d1 = d0 `mod` 146097 in
    let n100 = floor (fromIntegral d1 / 36524) in
    let d2 = d1 `mod` 36524 in
    let n4 = floor (fromIntegral d2 / 1461) in
    let d3 = d2 `mod` 1461 in
    let n1 = floor (fromIntegral d3 / 365) in
    let year = 400 * n400 + 100 * n100 + 4 * n4 + n1 in
    if n100 == 4 || n1 == 4 then
        year
    else
        year + 1

gregorianFromFixed :: Fixed -> Gregorian
gregorianFromFixed fd = 
    let year = gregorianYearFromFixed fd in
    let priorDays = fd - fixedFromGregorian (GregorianData year 1 1) in
    let correction = if fd < fixedFromGregorian (GregorianData year 3 1) then 
                         0 
                     else if fd >= fixedFromGregorian (GregorianData year 3 1) &&
                                   gregorianLeapYear year then 
                             1
                          else 
                             2
                 in
    let month = floor (fromIntegral ( 12 * (priorDays + correction) + 373) / 367) in
    let day = fd - fixedFromGregorian (GregorianData year month 1) + 1 in
    GregorianData year month day
                  