import Test.QuickCheck
import Text.Printf

import Calendar.Global
import Calendar.Julian
import Calendar.Gregorian
import Calendar.Islamic
import Calendar.Mayan
import Calendar.FrenchRevolutionary

detailed :: Config
detailed = Config
  { configMaxTest = 1000
  , configMaxFail = 10000
  , configSize    = (+ 3) . (`div` 2)
  , configEvery   = \n args -> let s = show n in s ++ [ '\b' | _ <- s ]
  }

main  = mapM_ (\(s,a) -> printf "%-25s: " s >> a) tests

-- The following cannot be used because Fixed is a type *synonym*
--instance Arbitrary Fixed where
--    arbitrary = choose (-100000, 10000)
--    coarbitrary = undefined
generateFixed :: Gen Fixed
generateFixed = do 
  i <- frequency [ (5, choose(10000,1000000)),
                   (3, choose(-1000000,-10000)),
                   (2, choose(-1000,1000)) ]
  return i

-- If we are close from the epoch (< 1000), we do not really risk a bug
-- We added during development a:
--          classify ((abs f) < 1000) "trivial" $
-- to check that our custom generator did not generate too many "trivial" values
-- Another way to test the distribution of test data is:
--          collect (floor (fromIntegral (abs f) / 100000)) $
-- to display nice statistics.
prop_julianRoundtrip = forAll generateFixed $ \f -> 
                             let j = julianFromFixed f in
                             ((fixedFromJulian . julianFromFixed) f == f &&
                             (julianFromFixed . fixedFromJulian) j == j)

prop_gregorianRoundtrip = forAll generateFixed $ \f -> 
                             let g = gregorianFromFixed f in
                             ((fixedFromGregorian . gregorianFromFixed) f == f &&
                             (gregorianFromFixed . fixedFromGregorian) g == g)

prop_gregorianNextDay = forAll generateFixed $ \f -> 
                        let nextFixed = f + 1 in
                        let gday = gregorianFromFixed f in
                        let nextGday = gregorianFromFixed nextFixed in
                        (
                         (  -- Same month
                           (gregorian'year gday == gregorian'year nextGday) &&
                           (gregorian'month gday == gregorian'month nextGday) &&
                           (gregorian'day gday + 1 == gregorian'day nextGday)
                         ) || (  -- Next month  
                           (gregorian'year gday == gregorian'year nextGday) &&
                           (gregorian'month gday +1 == gregorian'month nextGday) &&
                           (gregorian'day nextGday == 1)
                         ) || ( -- Next year
                           (gregorian'year gday + 1 == gregorian'year nextGday) &&
                           (gregorian'month nextGday == 1) &&
                           (gregorian'day nextGday == 1)
                         ))

prop_islamicRoundtrip = forAll generateFixed $ \f -> 
                        (fixedFromIslamic . islamicFromFixed) f == f

prop_mayanLongCountRoundtrip = forAll generateFixed $ \f -> 
                             let m = mayanLongCountFromFixed f in
                             ((fixedFromMayanLongCount . mayanLongCountFromFixed) f == f &&
                             (mayanLongCountFromFixed . fixedFromMayanLongCount) m == m)
-- There cannot be round trip tests for other Mayan calendars since they do not
-- include a year.

prop_frenchRevRoundtrip = forAll generateFixed $ \f -> 
                             let m = modifiedFrenchRevFromFixed f in
                             ((fixedFromModifiedFrenchRev . modifiedFrenchRevFromFixed) f == f &&
                             (modifiedFrenchRevFromFixed . fixedFromModifiedFrenchRev) m == m)

prop_frenchRevNextDay = forAll generateFixed $ \f -> 
                        let nextFixed = f + 1 in
                        let gday = modifiedFrenchRevFromFixed f in
                        let nextGday = modifiedFrenchRevFromFixed nextFixed in
                        (
                         (  -- Same month
                           (frenchrev'year gday == frenchrev'year nextGday) &&
                           (frenchrev'month gday == frenchrev'month nextGday) &&
                           (frenchrev'day gday + 1 == frenchrev'day nextGday)
                         ) || (  -- Next month  
                           (frenchrev'year gday == frenchrev'year nextGday) &&
                           (frenchrev'month gday + 1 == frenchrev'month nextGday) &&
                           (frenchrev'day nextGday == 1)
                         ) || ( -- Next year
                           (frenchrev'year gday + 1 == frenchrev'year nextGday) &&
                           (frenchrev'month nextGday == 1) &&
                           (frenchrev'day nextGday == 1)
                         ))

tests  = [
          ("Julian roundtrip", check detailed prop_julianRoundtrip),
          ("Gregorian roundtrip", check detailed prop_gregorianRoundtrip),
          ("Gregorian next day", check detailed prop_gregorianNextDay),
          ("Islamic roundtrip", check detailed prop_islamicRoundtrip),
          ("Mayan roundtrip", check detailed prop_mayanLongCountRoundtrip),
          ("French revolutionary roundtrip", check detailed prop_frenchRevRoundtrip),
          ("French revolutionary next day", check detailed prop_frenchRevNextDay)
         ]
